# Copyright 2022 neomadas-dev
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#
#   3. Neither the name of the copyright holder nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from django.template import Library, Node
from django.template.base import FilterExpression, Parser, Token
from django.template.loader_tags import construct_relative_path

register = Library()


class NeomImportNode(Node):
  def __init__(self, template: FilterExpression, *args, **kwargs):
    self.template = template
    super().__init__(*args, **kwargs)

  def render(self, context):
    subpath = self.template.resolve(context)
    fullpath = (
      construct_relative_path(self.origin.template_name, subpath),
    )

    cache = context.render_context.dicts[0].setdefault(self, {})
    template = cache.get(fullpath)

    if not template:
      template = context.template.engine.select_template(fullpath)
      cache[fullpath] = template

    return template.render(context)


@register.tag
def neom_import(parser: Parser, token: Token):
  bits = token.split_contents()
  if len(bits) < 2:
    raise template.TemplateSyntaxError(
      f'{bits[0]} tag takes at least one argument: the asset path'
    )
  bits[1] = construct_relative_path(parser.origin.template_name, bits[1])
  return NeomImportNode(parser.compile_filter(bits[1]))
